from transformers.generation.logits_process import (
    LogitsProcessorList,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)

MIN_TEMPERATURE = 1e-5
MIN_TOP_P = 1e-8


def prepare_logits_processor(
    temperature: float, repetition_penalty: float, top_p: float, top_k: int
) -> LogitsProcessorList:
    processor_list = LogitsProcessorList()
    # TemperatureLogitsWarper doesn't accept 0.0
    # 1.0 makes it a no-op so we skip two cases.
    if MIN_TEMPERATURE <= temperature < 1:
        processor_list.append(TemperatureLogitsWarper(temperature))
    if repetition_penalty > 1:
        processor_list.append(RepetitionPenaltyLogitsProcessor(repetition_penalty))
    if MIN_TOP_P <= top_p < 1:
        processor_list.append(TopPLogitsWarper(top_p))
    if top_k > 0:
        processor_list.append(TopKLogitsWarper(top_k))
    return processor_list


def partial_stop(output, stop_str):
    for i in range(0, min(len(output), len(stop_str))):
        if stop_str.startswith(output[-i:]):
            return True
    return False



import os
from datetime import datetime
from typing import Any, Dict, List

import torch  # type: ignore
from llmspec import (
    ChatChoice,
    ChatCompletionRequest,
    ChatMessage,
    CompletionResponse,
    Role,
    TokenUsage,
)
from mosec import Server, Worker, get_logger
from transformers import AutoModel, AutoTokenizer

logger = get_logger()
TOKENIZER = os.environ.get("MODELZ_TOKENIZER", "THUDM/chatglm-6b-int4")
MODEL = os.environ.get("MODELZ_MODEL", "THUDM/chatglm-6b-int4")


class Tokenizer(Worker):
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(
            TOKENIZER, trust_remote_code=True
        )

    def deserialize(self, buf: bytes) -> ChatCompletionRequest:
        return ChatCompletionRequest.from_bytes(buf)

    def forward(self, req: ChatCompletionRequest):
        prompt = req.get_prompt(MODEL)
        tokens = self.tokenizer(prompt, return_tensors="pt")
        # TODO: ignore the inference configurations for now
        return tokens.input_ids[0]


class Inference(Worker):
    def __init__(self):
        self.model = AutoModel.from_pretrained(MODEL, trust_remote_code=True)
        self.device = (
            torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
        )
        if torch.cuda.is_available():
            self.model = self.model.half().to(self.device)
        else:
            self.model = self.model.float()
        self.model.eval()

    def forward(self, tokens: List[Any]) -> Dict:
        inputs = torch.nn.utils.rnn.pad_sequence(tokens, batch_first=True).to(
            self.device
        )
        outputs = self.model.generate(inputs, max_length=30).tolist()
        return list(zip(tokens, outputs))


class Decoder(Worker):
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(
            TOKENIZER, trust_remote_code=True
        )

    def forward(self, data: Any):
        token, output = data
        res = output[len(token) :]
        msg = self.tokenizer.decode(res, skip_special_tokens=True)
        resp = CompletionResponse(
            id=MODEL,
            object="chat",
            created=datetime.now(),
            choices=[
                ChatChoice(message=ChatMessage(content=msg, role=Role.ASSISTANT)),
            ],
            usage=TokenUsage(
                prompt_tokens=len(token),
                completion_tokens=len(res),
                total_tokens=len(token) + len(res),
            ),
        )
        return resp

    def serialize(self, data: CompletionResponse):
        return data.to_json()


if __name__ == "__main__":
    server = Server()
    server.append_worker(Tokenizer, num=1, timeout=10)
    server.append_worker(Inference, num=1, max_batch_size=4, timeout=40)
    server.append_worker(Decoder, num=1, timeout=10)
    server.run()



import base64
from typing import List, Union

import numpy as np
import torch
import torch.nn.functional as F
import transformers
from llmspec import EmbeddingData, EmbeddingRequest, EmbeddingResponse, TokenUsage


class Emb:
    def __init__(self, model_name: str, device: str) -> None:
        self.model_name = model_name
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(model_name)
        self.model = transformers.AutoModel.from_pretrained(model_name)
        if device == "auto":
            self.device = (
                torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
            )
        else:
            self.device = device

        self.model = self.model.to(self.device)
        self.model.eval()

    # copied from https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2#usage-huggingface-transformers
    def get_embedding_with_token_count(self, sentences: Union[str, List[str]]):
        # Mean Pooling - Take attention mask into account for correct averaging
        def mean_pooling(model_output, attention_mask):
            # First element of model_output contains all token embeddings
            token_embeddings = model_output[0]
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                input_mask_expanded.sum(1), min=1e-9
            )

        # Tokenize sentences
        encoded_input = self.tokenizer(
            sentences, padding=True, truncation=True, return_tensors="pt"
        )
        inputs = encoded_input.to(self.device)
        token_count = inputs["attention_mask"].sum(dim=1).tolist()[0]
        # Compute token embeddings
        model_output = self.model(**inputs)
        # Perform pooling
        sentence_embeddings = mean_pooling(model_output, inputs["attention_mask"])
        # Normalize embeddings
        sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)

        return token_count, sentence_embeddings

    def __call__(self, req: EmbeddingRequest) -> EmbeddingResponse:
        token_count, embeddings = self.get_embedding_with_token_count(req.input)
        embeddings = embeddings.detach()
        if self.device != "cpu":
            embeddings = embeddings.cpu()
        embeddings = embeddings.numpy()
        if req.encoding_format == "base64":
            embeddings = [
                base64.b64encode(emb.astype(np.float32).tobytes()).decode("utf-8")
                for emb in embeddings
            ]
        else:
            embeddings = [emb.tolist() for emb in embeddings]

        resp = EmbeddingResponse(
            data=[
                EmbeddingData(embedding=emb, index=i)
                for i, emb in enumerate(embeddings)
            ],
            model=self.model_name,
            usage=TokenUsage(
                prompt_tokens=token_count,
                # No completions performed, only embeddings generated.
                completion_tokens=0,
                total_tokens=token_count,
            ),
        )
        return resp
